from Tkinter import *
import shapefile
import math

def projectMercatorSphere(point):
	# project point in latitude and longitude to Mercator x and y, using the formula for a sphere
	R = 6378206.4 # Clarke 1866 ellipsoid equatorial radius in meters
	lngDeg = point[0] # longitude of point in degrees
	latDeg = point[1] # latitude of point in degrees
	lambda0Deg = 0    # central meridian in degrees
	lambda0 = math.radians(lambda0Deg) # convert all 3 values to radians
	lmbda = math.radians(lngDeg)
	phi = math.radians(latDeg)
	# project radian coordinate values to Mercator x,y
	x = R * (lmbda - lambda0) # Mercator X coord
	y = R * math.log(math.tan(math.pi/4 + phi / 2.0)) # Mercator Y coord
	return (x,y)

def mercatorToScreen(mercXY, bboxProj):
	# At this point we have mercator x,y coords and a global bounding box in projected coords. We normalize
	# the projected coordinate pair mercXY to the y range of the projected bounding box for the shape and 
	# then create screen (pixel) coordinates by multiplying both normalized x and y by the height of the canvas.
	minProjX = bboxProj[0]
	minProjY = bboxProj[1]
	maxProjX = bboxProj[2]
	maxProjY = bboxProj[3]
	# normalize X and Y to the same coordinate range to keep 1:1 aspect ratio
	normY = 1 - ((mercXY[1] - minProjY) / (maxProjY - minProjY)) # Default Tk Y increases from top to bottom--invert for Mercator
	normX = (mercXY[0] - minProjX) / (maxProjY - minProjY)

	pixelX = normX * HEIGHT # create screen coords by multiplying both normalize X and Y 
	pixelY = normY * HEIGHT #   values by the same dimension (keep scaling constant in x and y)
	return (pixelX, pixelY) # returns a 2-tuple

def drawPart(startingPointIndex, endingPointIndex, shape, bboxProj, w):
	# project initial point in part to start the loop
	startPoint = shape.points[startingPointIndex]
	startMercXY = projectMercatorSphere(startPoint)
	startScreenXY = mercatorToScreen(startMercXY, bboxProj)
	firstScreenXY = startScreenXY # projected coords of first point in part--saved so that I can close the polygon
	
	for index in range(startingPointIndex+1, endingPointIndex):
		# This for loop starts at the second point in the shape and loops through to the last one. A line is
		# drawn from the projected values of the previous point to the projected values of this point. At the 
		# end of the loop, swap the current and previous projected points.
		point = shape.points[index]
		mercXY = projectMercatorSphere(point)
		screenXY = mercatorToScreen(mercXY, bboxProj)
		w.create_line(startScreenXY[0], startScreenXY[1], screenXY[0], screenXY[1]) # draw from previous to current projected point.
		startScreenXY = screenXY # swap current and previous point
	# close polygon
	w.create_line(startScreenXY[0], startScreenXY[1], firstScreenXY[0], firstScreenXY[1]) # close the polygon

def drawPolygon(shape, bboxProj, w):
	# draws a polygon shape, possibly with multiple parts (individual disconnected polygons)
	parts = shape.parts   # Starting indices of polygons in this shape, as list. Indexes are 0-based into shape.points.
	numParts = len(parts) # number of parts (polygons) in this shape
	part = 0              # starting with part 0...
	startingPointIndex = 0  # the index of the first point in a part of the shape
	endingPointIndex = 0    # the index of the last point in a part of the shape
	# Draw all but the last part in the shape.
	while part < numParts - 1:
	                      # I have created a while loop that allows me to read all but the last part of a shape. I do this
	                      # so that I can initialize startingPointIndex outside the while loop and then read a new
	                      # endingPointIndex for each part. Whenever I finish drawing a part, I copy endingPointIndex into
	                      # startingPointIndex. Then the loop starts again on the next shape. I handle the last part of the
	                      # shape outside of the while loop because its endingPointIndex is the number of points in the 
	                      # in the shape minus 1 (i.e., its not stated in the parts list)
		endingPointIndex = parts[part+1]
		drawPart(startingPointIndex, endingPointIndex, shape, bboxProj, w) # do the actual drawing on the widget here
		startingPointIndex = endingPointIndex	
		part += 1
	# draw last part in shape (its endingPointIndex is the number of points in the shape minus 1)
	startingPointIndex = endingPointIndex
	endingPointIndex = len(shape.points) - 1
	drawPart(startingPointIndex, endingPointIndex, shape, bboxProj, w)
	# print parts

def getShapefileBBox(shapes):
	# This returns a list that has the minimum and maximum coordinates for all the 
	# shapes in the shapefile. Each shape already knows its bounding box and is 
	# found in the property bbox.
	BBox = [40000000, 10000000, -40000000, -10000000] # minx, miny, maxx, maxy for all shapes (the global bounding box)
	for shape in shapes:               # Loop through all the shapes in the shapefile
		shapeBBox = shape.bbox         # Look at the bounding box for each shape and 
		#print shapeBBox               # see if it extends the global bounding box.
		if shapeBBox[0] < BBox[0]:     # If it does, update the current bounding box.
			BBox[0] = shapeBBox[0]
		if shapeBBox[1] < BBox[1]:
			BBox[1] = shapeBBox[1]
		if shapeBBox[2] > BBox[2]:
			BBox[2] = shapeBBox[2]
		if shapeBBox[3] > BBox[3]:
			BBox[3] = shapeBBox[3]
	return BBox

def projectBoundingBox(bbox):
	# Take bbox (assumed to be in lat/long), project it, and return the projected coords
	lowerLeft = projectMercatorSphere((bbox[0], bbox[1]))
	upperRight = projectMercatorSphere((bbox[2], bbox[3]))
	return [lowerLeft[0], lowerLeft[1], upperRight[0], upperRight[1]]

def drawAllShapes():
	# draws all the shapes in the shapefile
	# Right now it only draws polygon shapes (no polylines, etc)
	bbox = getShapefileBBox(sf.shapes()) # Find the minimum bounding rectangle (MBR) in native coords for the shapefile (allows 'zoom to extent')
	print 'bbox lat long:', bbox

	bboxProj = projectBoundingBox(bbox) # Find the projected values for the MBR (in our case using Mercator, spherical case)
	print 'bbox projected: ', bboxProj

	shapes = sf.shapes()           # the list of shapes in the shapefile

	for shape in shapes:           # for each shape in the shapefile, 
		points = shape.points      # access the list of points in the shape
		if shape.shapeType == 5:   # if it's a polygon... (see ESRI shapefile specification(https://www.esri.com/library/whitepapers/pdfs/shapefile.pdf)
			drawPolygon(shape, bboxProj, w) # draw the points in the polygon

def configure(event):
	pass
	# This is a callback for the <Configure> event that fires when a widget is resized.
	# Anything in here would be run when the user resizes the widget. All the commented stuff below works but is unnecessary for this app.
	# print 'in configure'
	# print event.width, event.height
	# canvas = event.widget
	# drawAllShapes() # not necessary--once I created a <Configure> event handler the canvas widget automatically redrew correctly on resizing

if __name__ == '__main__':

	# This is the global namespace for the program. Variables created here are visible to any function called from here!!!

	master = Tk() # create an empty frame (the 'master' frame)
	WIDTH = 800   # start out with these pixel values for width and height of the Canvas widget
	HEIGHT = 400
	w = Canvas(master, width=WIDTH, height=HEIGHT) # create a canvas widget attached to master
	                                               # http://effbot.org/tkinterbook/canvas.htm
	w.bind("<Configure>", configure)               # create a window resize event handler (handles the <Configure> event)
	w.pack()                                       # make the canvas fill the master frame

	# Uncomment one of these lines to read in a shapefile from the shapefiles folder
	# The shapefile folder is assumed to be in the same folder as this module
	# sf = shapefile.Reader('shapefiles/blockgroups')
	# sf = shapefile.Reader('shapefiles/states')
	sf = shapefile.Reader('shapefiles/surf2g')
	
	# Since sf is defined in the global namespace, it is visible to drawAllShapes(). This allows drawAllShapes() to be called from the <Configure> event
	# handle configure() (although that doesn't seem necessary)
	drawAllShapes()

	mainloop() # start the event handling loop

